freeslot(
	"SPR_ELSP",
	"SPR_SPXT",
	"S_SPARKSPINSPARK",
	"S_SPARKSPINSPARK_THRUST",
	"MT_SPARKSPINSPARK",
	"S_SPARKS_CHARGE",
	"MT_SPARKS_CHARGE",
	"SPR2_PROP",
	"SPR2_DROP",
	"SPR2_WAI2",
	"SPR2_FKIK",
	"SKINCOLOR_ELLIE"
)

skincolors[SKINCOLOR_ELLIE] = {
        name = 'Ellie',
        ramp = {177,178,179,180,181,182,163,164,164,164,165,165,166,168,169,31},
        invcolor = SKINCOLOR_PASTEL,
        invshade = 3,
        chatcolor = V_MAGENTAMAP,
        accessible = true
}


if not kirbyabilitytable then
	rawset(_G, 'kirbyabilitytable', {})
end
kirbyabilitytable['sparks'] = 14

states[S_SPARKSPINSPARK] = {
	sprite = SPR_ELSP,
	frame = A|FF_FULLBRIGHT|FF_ANIMATE|FF_RANDOMANIM,
	var1 = 3,
	var2 = 2,
	duration = -1
}

mobjinfo[MT_SPARKSPINSPARK] = {
	spawnstate = S_SPARKSPINSPARK,
	flags = MF_NOCLIP | MF_NOGRAVITY | MF_SCENERY,
	height = 4*FRACUNIT
}

addHook("MobjThinker", function(mo)
	if mo.doGhost and (leveltime & 1) then
		--disabled for performance
		--P_SpawnGhostMobj(mo)
	end
	mo.momz = $+FRACUNIT/8
end, MT_SPARKSPINSPARK)

local TWIRL_FRAMES = 8                -- Number of frames in the twirl (SPR2_PROP) animation.
local TWIRL_LENGTH = 25               -- The amount of frames a twirl lasts after being released.
local TWIRL_HOLD_WEIGHT = 12*FRACUNIT  -- While holding a twirl, Sparks will move down by this much each frame.
local TWIRL_HOLD_WEIGHT_CHARGED = 6*FRACUNIT

local TWIRL_MOMZ_RESTORE_MULTIPLIER = 0 --FRACUNIT/10
local TWIRL_MOMZ_BOOST_BASE = 20*FRACUNIT
local TWIRL_MOMZ_BOOST_PER_SEC = 20*FRACUNIT
local TWIRL_MOMZ_BOOST_MAX = 40*FRACUNIT 
local TWIRL_RADIUS = 48*FRACUNIT
local TWIRL_RADIUS_CHARGED = 192*FRACUNIT

local DROP_MOMZ = 25*FRACUNIT         -- The speed of the drop attack.
local DROP_DUST_SPAWN_RADIUS = 4
local DROP_DUST_SPEED = 4
local DROP_ATTACK_RADIUS = 192*FRACUNIT

local TWIRL_WITHOUT_GLIDE = true		-- If true, pushing Spin in mid-air when not gliding will do a twirl attack if possible, or
                                        -- a drop attack if not. If false, pushing Spin when gliding will always do a drop attack.
										-- Pushing Spin while gliding will never do a drop attack no matter what this is set to.
										-- I think False is better because it makes the behavior of Spin have a consistent effect,
										-- and makes the drop attack always accessible, but I've left this here in case I change my
										-- mind, or you want to change it yourself. I might make this configurable via a console
										-- command in a later version.
										-- ...is what I was *going* to say, but then I discovered that it becomes really easy to spam
										-- 1-frame "drop" attacks by holding Spin and jumping. Not sure if this is exploitable or not,
										-- but it seems rather messy.

local TWIRL_HOLD_SOUND = sfx_s3ka0      -- The sound to play while holding a twirl. (sfx_s25e also works)
local TWIRL_HOLD_SOUND_VOL = 255        -- The volume at which to play the hold sound. (128 works for sfx_s25e)
local TWIRL_HOLD_SOUND_CHARGED = sfx_s3k78
local TWIRL_RELEASE_SOUND = sfx_drill1  -- The sound to play upon releasing a twirl.
local TWIRL_RELEASE_SOUND_CHARGED = sfx_s3k5e
local DROP_START_SOUND = sfx_trfire
local DROP_GROUND_SOUND = sfx_pstop     -- The sound to play upon hitting the ground at the end of a drop attack.
local DROP_START_SOUND_CHARGED = sfx_s3k45
local DROP_GROUND_SOUND_CHARGED = sfx_s3k40

local GLIDE_UNDERWATER_START_SOUND = sfx_splash
local GLIDE_UNDERWATER_LOOP = sfx_amwtr6
local GLIDE_UNDERWATER_LOOP_VOL = 70

local ELEC_DRAIN_COUNT = 32
local ELEC_CHARGE_SECS = 30
local ELEC_DRAIN_TICS = ELEC_CHARGE_SECS*TICRATE/ELEC_DRAIN_COUNT
local ELEC_LOW_THRESHOLD = 6
local ELEC_CHARGE_SOUND = sfx_buzz1

local ROLL_START_SOUND = sfx_s3k47
local ROLL_STOP_SOUND = sfx_s3k77

local WAI2_TIME = 5*TICRATE

-- player.twirl_state:
--  0: Sparks is not twirling, but has the attack ready.
--  1: Sparks is currently holding a twirl (or has *just* released it)
--  2..TWIRL_LENGTH: Sparks has released a twirl, and is still spinning.
--  TWIRL_LENGTH+1..: Sparks has already used her twirl.

-- player.twirl_frame:
--  The number of frames since the player has last started a twirl. Increments even while holding it.

-- player.drop_state:
--  0: Sparks is not using her drop attack.
--  1: Sparks is using her drop attack.
--  2: Sparks is unable to use her drop attack because she was hit.

local function should_override_jump_spin(player)
	return player.rings < 50 or not All7Emeralds(emeralds) or player.powers[pw_shield] ~= SH_NONE or player.powers[pw_super]
end

local function spin_button_starts_twirl(player)
	-- Condition for whether or not a twirl should start upon pressing the spin button.
	if player.twirl_state ~= 0 then
		return false
	elseif player.secondjump then
		return true
	elseif P_IsObjectOnGround(player.mo) then
		return false
	else
		return should_override_jump_spin(player)
	end
end

local function start_drop(player)
	if player.drop_state == 0 and not P_IsObjectInGoop(player.mo) then
		player.drop_state = 1
		player.suppress_spin = true  -- don't spin when hitting the ground unless the player presses the button a second time
		S_StartSound(player.mo, DROP_START_SOUND)
		if player.electricity then
			S_StartSound(player.mo, DROP_START_SOUND_CHARGED)
		end
	end
end

local function start_twirl(player)
	if player.twirl_state == 0 then
		player.twirl_frame = 0
	end
	player.twirl_state = 1
	player.drop_state = 0
	if player.saved_momz == nil then
		-- Save our current momentum to add onto upon releasing
		player.saved_momz = player.mo.momz
	end
end

local function sparkle(player, x, y, z, fuse, color)
	if x == nil then
		for i=1,2 do
			local h_radius = 16
			local v_extent = 32
			local speed = y
			local x = P_RandomRange(-h_radius, h_radius) * FRACUNIT
			local y = P_RandomRange(-h_radius, h_radius) * FRACUNIT
			local z = P_RandomRange(0, v_extent) * FRACUNIT
			local sp = sparkle(player, x, y, z, 2*TICRATE)
			if speed ~= nil then
				sp.momx = speed * x
				sp.momy = speed * y
				sp.momz = speed * z
			end
		end
	else
		local mo = P_SpawnMobjFromMobj(player.mo, x, y, z, MT_SPARKSPINSPARK)
		mo.color = color or player.mo.color
		mo.fuse = fuse
		if player.electricity then
			mo.doGhost = true
		end
		return mo
	end
end

local function elec_fx_color(player)
	if player.powers[pw_shield] & SH_PROTECTELECTRIC then
		return SKINCOLOR_YELLOW
	else
		return player.mo.color
	end
end

local function elec_sparkle(player, radius, count)
	for i=1,count do
		local angle = ANG1*P_RandomRange(0, 359)
		local r = FixedMul(radius, P_RandomFixed())
		local x = FixedMul(cos(angle), r)
		local y = FixedMul(sin(angle), r)
		local z = P_RandomRange(0, player.mo.height/FRACUNIT) * FRACUNIT
		sparkle(player, x, y, z, TICRATE/4, elec_fx_color(player))
	end
end

local function elec_always_full(player)
	return player.powers[pw_super] or (player.powers[pw_shield] & SH_PROTECTELECTRIC)
end

addHook("PreThinkFrame", function()
	for player in players.iterate do
		if not player.valid or not player.mo or not player.mo.valid or player.mo.skin ~= 'sparks' then
			continue
		end
		
		if (player.cmd.buttons & BT_JUMP) and not (player.pflags & PF_JUMPDOWN) then
			player.suppress_spin = true
		end
		
		if not (player.cmd.buttons & BT_SPIN) then
			-- The Spin button was released, so the next time is probably an intentional spin.
			-- Unless it's a twirl or a drop, but in the latter case the variable will be set back to true.
			player.suppress_spin = false
		elseif player.suppress_spin then
			player.cmd.buttons = $ & ~BT_SPIN
		end
	end
end)

local function init_player(player)
	if player.valid and player.mo and player.mo.valid and player.mo.skin == 'sparks' then
		player.twirl_state = 0
		player.gliding_frames = 0
		player.electricity = 0
		player.suppress_spin = false
	end
end

addHook("MapLoad", function()
	for player in players.iterate do
		init_player(player)
	end
end)

-- Effect 1 ("Slope Skew"): Don't play sound
addHook('LinedefExecute', function(line, mo)
	if mo.valid and mo.player and mo.player.valid and mo.skin == 'sparks' then
		mo.player.electricity = ELEC_DRAIN_COUNT
		if not (line.flags & ML_EFFECT1) then
			S_StartSound(mo, ELEC_CHARGE_SOUND)
		end
	end
end, 'CHARGE_SPARKS_NO_STUN')

-- Effect 1 ("Slope Skew"): Don't play sound
addHook('LinedefExecute', function(line, mo)
	if mo.valid and mo.player and mo.player.valid and mo.skin == 'sparks' then
		if mo.player.electricity > 0 and not (line.flags & ML_EFFECT1) then
			S_StartSound(mo, sfx_adderr)
		end
		mo.player.electricity = 0
	end
end, 'DISCHARGE_SPARKS')

-- Front Texture X: Tag to execute if Sparks is charged
-- Front Texture Y: Tag to execute if Sparks is not charged
-- Back Texture X:  Tag to execute if the activator is not Sparks
addHook('LinedefExecute', function(line, mo, sector)
	if not (mo.valid and mo.player and mo.player.valid) then return end
	
	local trigger = function(offset)
		if offset then
			P_LinedefExecute(offset/FRACUNIT, mo, sector)
		end
	end
	
	if mo.skin == 'sparks' then
		if mo.player.electricity then
			trigger(line.frontside.textureoffset)
		else
			trigger(line.frontside.rowoffset)
		end
	elseif line.backside and line.backside.valid then
		trigger(line.backside.textureoffset)
	end
end, 'CHECK_SPARKS_ELEC_STATUS')

addHook("ThinkFrame", function()
	for player in players.iterate do
		if not player.valid or player.mo == nil or player.mo.skin ~= 'sparks' then
			continue
		end
		
		if not (player.cmd.buttons & BT_SPIN) then
			--failsafe in case it gets stuck
			player.suppress_spin = false
		end
		
		if player.mo.sprite2 == SPR2_WAIT or player.mo.sprite2 == SPR2_WAI2 then
			player.wait_frames = ($ or 0) + 1
			if player.wait_frames >= WAI2_TIME then
				player.mo.sprite2 = SPR2_WAI2
				player.mo.frame = (player.wait_frames / 8) % 2
			end
		else
			player.wait_frames = 0
		end
		
		if player.twirl_state == nil or player.playerstate == PST_DEAD then
			init_player(player)
		elseif player.twirl_state > 1 or (player.twirl_state == 1 and not (player.cmd.buttons & BT_SPIN)) then
			-- If we're twirling, and the player has already released the button, then increment twirl_state
			player.twirl_state = $+1
		end
		-- Increment twirl_frame
		player.twirl_frame = ($ or 0) + 1
		
		if player.twirl_state > 0 and player.twirl_state <= TWIRL_LENGTH then
			if player.electricity then
				elec_sparkle(player, FixedMul(player.mo.scale, TWIRL_RADIUS_CHARGED), 16)
			else
				elec_sparkle(player, FixedMul(player.mo.scale, TWIRL_RADIUS), 8)
			end
		
			if player.twirl_state == 1 then  -- If we are currently holding a twirl...
				-- ...move the player toward the ground faster, since Sparks isn't supposed to be able to glide very far while attacking.
				local weight
				if player.electricity then
					weight = TWIRL_HOLD_WEIGHT_CHARGED
				else
					weight = TWIRL_HOLD_WEIGHT
				end
				weight = FixedMul($, player.mo.scale)
				player.mo.z = $-weight*P_MobjFlip(player.mo)
			else
				S_StopSoundByID(player.mo, TWIRL_HOLD_SOUND) -- Stop the hold sound
				--if (player.cmd.buttons & BT_SPIN) and not (player.cmd.buttons & BT_JUMP) and player.twirl_state then
					-- If we've already released a twirl, and the player has released jump and pressed spin, trigger the drop move.
					-- moved to JumpSpinSpecial
					--start_drop(player)
				--end
			end
		end
		
		if elec_always_full(player) then
			player.electricity = ELEC_DRAIN_COUNT
		elseif leveltime % ELEC_DRAIN_TICS == 0 and (player.electricity or (player.elfride and player.elfride.elf)) then
			if player.elfride and player.elfride.elf then
				player.electricity = min($+1, ELEC_DRAIN_COUNT)
			else
				player.electricity = $-1
			end
		end
		
		if player.electricity then
			local avg_spark_time = TICRATE
			if elec_always_full(player) then
				avg_spark_time = 2*$
			end
			local spark_chance
			if (player.twirl_state > 0 and player.twirl_state <= TWIRL_LENGTH) or player.drop_state == 1 then
				spark_chance = FRACUNIT/4
			else
				spark_chance = min(player.electricity, ELEC_LOW_THRESHOLD) * FRACUNIT/ELEC_LOW_THRESHOLD/avg_spark_time
			end
			if P_RandomChance(spark_chance) then
				local zap = P_SpawnMobjFromMobj(player.mo, 0, 0, 0, MT_SPARKS_CHARGE)
				zap.color = elec_fx_color(player)
				zap.colorized = true
				if player.powers[pw_super] then
					zap.scale = player.mo.scale
				else
					zap.scale = player.mo.scale * 2 / 3
				end
			end
			
			--if player.mo.eflags & MFE_UNDERWATER and not player.powers[pw_super] then
				--player.electricity = 0
			--end
		end
		
		if player.force_bust_floor then
			-- To ensure the bustable floor check still works on the frame where Sparks hits the ground,
			-- despite drop_state being set to 0 on that frame.
			player.force_bust_floor = false
		end
		
		if P_IsObjectOnGround(player.mo) then
			if player.drop_state == 1 then
				-- If we were dropping, and we've hit the ground, finish the move.
				S_StartSound(player.mo, DROP_GROUND_SOUND)
				if player.electricity then
					S_StartSound(player.mo, DROP_GROUND_SOUND_CHARGED)
					elec_sparkle(player, FixedMul(player.mo.scale, DROP_ATTACK_RADIUS), 128)
					for i=0,35 do
						local x = cos(ANG10*i)
						local y = sin(ANG10*i)
						local mo = sparkle(player, 4*x, 4*y, 0, TICRATE/2, SKINCOLOR_YELLOW)
						mo.momx = FixedMul(x, DROP_ATTACK_RADIUS-4*player.mo.scale) * 2
						mo.momy = FixedMul(y, DROP_ATTACK_RADIUS-4*player.mo.scale) * 2
					end
				end
				for i=0,17 do
					local a = i*ANG1*20
					local x = cos(a)
					local y = sin(a)
					local adj_radius = FixedMul(player.mo.scale, DROP_DUST_SPAWN_RADIUS)
					local dust = P_SpawnMobjFromMobj(player.mo, x*adj_radius, y*adj_radius, 4*FRACUNIT, MT_DUST)
					--dust.color = player.mo.color
					--dust.colorized = true
					dust.momx = DROP_DUST_SPEED * x
					dust.momy = DROP_DUST_SPEED * y
				end
				if player.electricity then
					P_RadiusAttack(player.mo, player.mo, FixedMul(player.mo.scale, DROP_ATTACK_RADIUS), DMG_ELECTRIC)
				end
				player.force_bust_floor = true
			end
			-- Reset drop_state since we're on the ground, whether it's 1 or 2.
			player.drop_state = 0
		end
		
		if player.mo.eflags & MFE_SPRUNG and player.drop_state == 1 then
			for i,v in ipairs({'momx', 'momy', 'momz'}) do
				player.mo[v] = $ * 3/2
			end
			S_StartSound(player.mo, sfx_sprong)
			player.drop_state = 0
		elseif P_IsObjectInGoop(player.mo) then
			-- Abort the drop attack if we're in goop; the player would otherwise get stuck
			-- since P_IsObjectOnGround would always return false.
			player.drop_state = 0
		end
		
		-- Player state implementations
		if player.drop_state == 1 then
			-- Drop attack
			player.mo.sprite2 = SPR2_DROP
			player.mo.anim_duration = 1
			player.mo.frame = leveltime % 12
			player.mo.momz = FixedMul(player.mo.scale, -DROP_MOMZ*P_MobjFlip(player.mo))
			player.saved_momz = player.mo.momz
			--P_SpawnMobjFromMobj(player.mo, 0, 0, 0, MT_SUPERSPARK)
			sparkle(player)
		elseif player.twirl_state > 0 and player.twirl_state <= TWIRL_LENGTH and not P_IsObjectOnGround(player.mo) then
			-- Twirl attack
			player.mo.sprite2 = SPR2_PROP
			player.mo.frame = player.twirl_frame % TWIRL_FRAMES
			player.mo.anim_duration = 1
			if player.electricity then
				P_RadiusAttack(player.mo, player.mo, FixedMul(TWIRL_RADIUS_CHARGED, player.mo.scale), DMG_ELECTRIC)
			else
				P_RadiusAttack(player.mo, player.mo, FixedMul(TWIRL_RADIUS, player.mo.scale), DMG_ELECTRIC)
			end
			if player.twirl_state == 1 then
				-- Still holding
				if not S_SoundPlaying(player.mo, TWIRL_HOLD_SOUND) then
					S_StartSoundAtVolume(player.mo, TWIRL_HOLD_SOUND, TWIRL_HOLD_SOUND_VOL)
				end
				if player.electricity then
					if not S_SoundPlaying(player.mo, TWIRL_HOLD_SOUND_CHARGED) then
						S_StartSoundAtVolume(player.mo, TWIRL_HOLD_SOUND_CHARGED, TWIRL_HOLD_SOUND_VOL)
					end
				end
			elseif player.twirl_state == 2 then
				-- Just released
				S_StartSound(player.mo, TWIRL_RELEASE_SOUND)
				if player.electricity then
					S_StartSound(player.mo, TWIRL_RELEASE_SOUND_CHARGED)
				end
				local boost = TWIRL_MOMZ_BOOST_BASE + TWIRL_MOMZ_BOOST_PER_SEC * player.twirl_frame / TICRATE
				boost = min($, TWIRL_MOMZ_BOOST_MAX)
				local flip = P_MobjFlip(player.mo)
				local new_momz = max(0, FixedMul(player.saved_momz, TWIRL_MOMZ_RESTORE_MULTIPLIER))
				new_momz = $ + FixedMul(boost * P_MobjFlip(player.mo), player.mo.scale)
				if flip == -1 then
					-- Gravity is reversed, so take the lower of the old momentum and the new momentum
					player.mo.momz = min(new_momz, player.saved_momz)
				else
					-- Gravity is normal, so take the higher of the two
					player.mo.momz = max(new_momz, player.saved_momz)
				end
				
				player.saved_momz = nil
			end
			sparkle(player)
		elseif player.dashmode > 3*TICRATE then
			-- Dash mode
			if player.mo.sprite2 ~= SPR2_DASH then
				-- Make sure it's always using the dash mode sprite
				player.mo.frame = $ % 2
			end
			player.mo.sprite2 = SPR2_DASH
			if player.secondjump then
				-- If we're gliding at the same time, leave a trail of glitter
				P_SpawnMobjFromMobj(player.mo, 0, 0, 0, MT_NIGHTSPARKLE)
			end
		elseif player.secondjump then
			-- If we're gliding, and none of the above cases apply, then make sure it's using the float sprite.
			if player.mo.eflags & MFE_UNDERWATER then
				player.mo.sprite2 = SPR2_SWIM
				player.mo.frame = leveltime % 16
				player.mo.anim_duration = 8
				if player.gliding_frames == 0 then
					S_StartSound(player.mo, GLIDE_UNDERWATER_START_SOUND)
				end
				if not S_SoundPlaying(player.mo, GLIDE_UNDERWATER_LOOP) then
					S_StartSoundAtVolume(player.mo, GLIDE_UNDERWATER_LOOP, GLIDE_UNDERWATER_LOOP_VOL)
				end
				if not P_IsObjectInGoop(player.mo) then
					local min_momz = min(0, FixedHypot(player.mo.momx, player.mo.momy) - 8*player.mo.scale)
					player.mo.momz = max(min_momz, $)
				end
			else
				-- Normally, the run animation would override it; this code will make sure that doesn't happen.
				-- Only displaying the first frame in that case is intentional, though it wasn't at first.
				player.mo.frame = $ % 2
				player.mo.sprite2 = SPR2_FLT_
			end
			player.gliding_frames = $ + 1
		end
		
		if player.pflags & PF_SPINNING or player.powers[pw_super] then
			sparkle(player)
		end
		
		if not player.secondjump then
			player.gliding_frames = 0
			S_StopSoundByID(player.mo, GLIDE_UNDERWATER_LOOP)
		end
		
		if P_IsObjectOnGround(player.mo) then
			-- If we're on the ground, reset the twirl attack.
			player.twirl_state = 0
			player.allow_another_twirl = false
		elseif (player.cmd.buttons & BT_SPIN) and player.secondjump and spin_button_starts_twirl(player) then
			-- Start the twirl attack if appropriate.
			-- Due to the "player.secondjump" check, this will only take effect while gliding, and is here because
			-- "JumpSpinSpecial" does not trigger while gliding. The code to start/hold a twirl when *not* gliding
			-- (not possible if TWIRL_WITHOUT_GLIDE is false) is in the "JumpSpinSpecial" handler below.
			start_twirl(player)
		elseif (player.allow_another_twirl or player.powers[pw_super]) and player.twirl_state > TWIRL_LENGTH then
			-- We've finished a twirl attack. It's normally only once per jump, so we don't usually reset it until hitting the ground.
			-- However, there are two conditions in which we do want the player to be able to twirl again:
			--  1) The player has hit an enemy in mid-air (which sets the "allow_another_twirl" flag), or
			--  2) The player is in Super form (in which case it's unlimited.)
			-- So, in these cases, we'll reset it.
			player.allow_another_twirl = false
			player.twirl_state = 0
		end
		
		if player.lastspr2 == SPR2_ROLL then
			if player.mo.sprite2 ~= SPR2_ROLL then
				S_StartSound(player.mo, ROLL_STOP_SOUND)
				for a=0,50 do
					sparkle(player, nil, 2)
				end
			end
		elseif player.mo.sprite2 == SPR2_ROLL and player.lastspr2 ~= SPR2_ROLL then
			S_StartSound(player.mo, ROLL_START_SOUND)
		end
		player.lastspr2 = player.mo.sprite2
	end
end)

addHook("PlayerCanDamage", function(player)
	if player.valid and player.mo.skin == 'sparks'  -- If this is a valid player, playing as Sparks...
	and ((player.twirl_state and player.twirl_state <= TWIRL_LENGTH) or player.drop_state==1) then  -- ...and the player is either twirling or dropping...
		return true   -- ...then Sparks will attack enemies.
	end
	-- Otherwise, the function will return nil and the normal game logic will apply.
end)

local function mobjIsSparks(mo)
	-- Is the given mobj Sparks' player object?
	if mo == nil then
		return false
	else
		return mo.player ~= nil and mo.skin == 'sparks'
	end
end

addHook("MobjDamage", function(target, inflictor, source, damage, damagetype)
	if mobjIsSparks(target) then
		local player = target.player
		-- If Sparks *was* damaged, then prevent her from twirling or dropping until she reaches the ground.
		player.twirl_state = TWIRL_LENGTH + 1
		player.drop_state = 2
		
		if damagetype == DMG_ELECTRIC then
			local elec = player.electricity
			player.electricity = ELEC_DRAIN_COUNT
			S_StartSound(target, ELEC_CHARGE_SOUND)
			P_DoPlayerPain(player, source, inflictor)
			return true
		end
	end
	if mobjIsSparks(inflictor) then
		-- If Sparks was the one who *dealt* the damage, then set the allow_another_twirl flag.
		-- No need to check whether she's on the ground, as the flag will just be immediately cleared in that case.
		inflictor.player.allow_another_twirl = true
	end
end)

addHook("ShouldDamage", function(target, inflictor, source, damage, damagetype)
	if mobjIsSparks(target) then
		if inflictor == nil and source == nil and damagetype == DMG_FIRE and target.player.dashmode > 3*TICRATE then
			-- Make Sparks immune to lava while in dash mode, because riding her scarf-board over lava is awesome.
			-- Plus it makes for interesting gameplay: if she slows down on lava, she'll exit dash mode and take damage.
			return false
		end
	end
end)

addHook("JumpSpinSpecial", function(player)
	if player.mo.skin == 'sparks' then
		if TWIRL_WITHOUT_GLIDE then
			if spin_button_starts_twirl(player) then
				start_twirl(player)
				return true
			elseif player.twirl_state > 1 and should_override_jump_spin(player) then
				start_drop(player)
				return true
			end
		elseif should_override_jump_spin(player) and player.twirl_state ~= 1 then
			start_drop(player)
			return true
		end
	end
end)

addHook("SpinSpecial", function(player)
	if player.suppress_spin and player.mo.skin == 'sparks' then
		return true
	end
end)

addHook("AbilitySpecial", function(player)
	-- Prevent glide while dropping
	return player.mo.skin == 'sparks' and player.drop_state == 1
end)

-- Based on Tatsuru's code on SRB2 Discord
local function CheckAndCrumble(mo, sec)
  for fof in sec.ffloors()
    if not (fof.flags & FF_EXISTS) continue end -- Does it exist?
    if not (fof.flags & FF_BUSTUP) continue end -- Is it bustable?
    
    if mo.z + mo.height < fof.bottomheight continue end -- Are we too low?
    if mo.z > fof.topheight continue end -- Are we too high?

    -- Check for whatever else you may want to

    EV_CrumbleChain(fof) -- Crumble
  end
end

addHook("PlayerThink", function(p)
	if not p.mo return end
	if p.mo.skin != "sparks" return end
	if p.drop_state ~= 1 and not p.force_bust_floor return end
	CheckAndCrumble(p.mo, p.mo.subsector.sector)
end)
-- End Tatsuru's code

hud.add(function(v, player)
	if player.mo == nil then
		return
	end
	local target_player = player
	if player.mo.skin == 'elfilin' then
		if player.elfilin and player.elfilin.ridingplayer then
			target_player = player.elfilin.ridingplayer.player
		end
	end
	if target_player.mo.skin ~= 'sparks' then
		return
	end
	
	local chars = '0123456789ABCDEFGHIJKLMNOPQRSTUV'
	if target_player.electricity > ELEC_LOW_THRESHOLD or (leveltime & 1 and target_player.electricity) or (target_player.elfride and target_player.elfride.elf) then
		local cmap = v.getColormap(target_player.mo.skin, target_player.mo.color)
		local patch = v.cachePatch('SPRXPWR'..chars:sub(target_player.electricity, target_player.electricity))
		local flags = V_PERPLAYER|V_MEDSCALEPATCH
		if elec_always_full(target_player) then
			flags = $ | V_HUDTRANS
		else
			flags = $ | V_HUDTRANSDOUBLE
		end
		v.draw(17, 200, patch, flags, cmap)
	end
end, 'game')